-- myConv.e
-- v1.1
-- 2012-09-09
-- conversion functions
-- adapted to Euphoria 4.1

-- function dec2str(atom n, atom d)
-- function base2str(atom n, atom b, atom d)
-- function num2str(atom n, atom b, atom d)
-- function str2num(sequence s, atom b)
-- function byte2word(atom l, atom h)
-- function highByte(atom x)
-- function lowByte(atom x)
-- function val(sequence str)
--   returns numerical value of string str
--   decimal separator may be a comma
--   displays a warning on error
-- function atom2string(atom val, sequence fmt)
--   converts numerical value val according to format fmt
--   replaces decimal separator by a comma (for non english speaking countries)
-- function object2string(object s)
--   converts an object in a sequence
--   sub-sequences appear between curly brackets as with print()
--   sub-sequence items are separated by commas as with print()
-- function utf8ToAscii(sequence s)
-- function hexString(sequence s)
--   returns hexadecimal representation of a sequence
-- function dos2unix(sequence s)
--   converts DOS end of lines to Unix ones

include std/types.e
include std/get.e
include std/convert.e
include std/sequence.e
include myTypes.e
include myDebug.e

public constant
  NORMAL = 1, REVERSE = 2

-------------------------------------------------------------------------------

public function dec2str(atom n, atom d)
  sequence result

  result = repeat(32, d)
  for i = d to 1 by -1 do
    result[i] = remainder(n, 10) + 48
    n = floor(n / 10)
  end for
  return result
end function

-------------------------------------------------------------------------------

public function base2str(atom n, atom b, atom d)
  atom c
  sequence result

  result = repeat(32, d)
  c = floor((log(b)/log(2))+0.5)
  for i = d to 1 by -1 do
    result[i] = DIGITS[and_bits(n / power(2, (d-i)*c), b-1)+1]
  end for
  return result
end function

-------------------------------------------------------------------------------

public function num2str(atom n, atom b, atom d)

  if b = 2 then
    if d>32 then d = 32 end if
    return base2str(n, 2, d)
  elsif b = 8 then
    if d>11 then d = 11 end if
    return base2str(n, 8, d)
  elsif b = 10 then
    return dec2str(n, d)
  elsif b = 16 then
    if d>8 then d = 8 end if
    return base2str(n, 16, d)
  else
    return {}
  end if
end function

-------------------------------------------------------------------------------

public function str2num(sequence s, atom b)
  atom l

  l = 0
  for i = 1 to length(s) do
    l = l + (match({s[i]}, DIGITS)-1) * floor(power(b, length(s)-i))
  end for
  return l
end function

-------------------------------------------------------------------------------

-- joinBits(higher_bits, lower_bits, divide)
-- combine higher and lower item to build one bigger item
-- divide is the resulting value when highest is 1 and lowest is 0
-- examples:
--  joinBits(h, l, #10) joins to quartets (4 bits) to build a byte (8 bits)
--  joinBits(h, l, #100) joins to bytes (8 bits) to build a word (16 bits)
--  joinBits(h, l, #10000) joins to words (16 bits) to build a dword (32 bits)
public function joinBits(integer higher_bits, integer lower_bits, integer divide)
  return (higher_bits*divide) + lower_bits
end function

-------------------------------------------------------------------------------

-- higherBits(x, divide)
-- returns the highest bits of an integer
-- divide is the is the value that fixes the limit between higher bits and lower bits
-- examples:
--  higherBits(x, #10) gets the higher quartet (4 bits) of byte x (8 bits)
--  higherBits(x, #100) gets the higher bytes (8 bits) of word x (16 bits)
--  higherBits(x, #10000) gets the higher word (16 bits) of dword x (32 bits)
public function higherBits(integer x, integer divide)
  return floor(x / divide)
end function

-------------------------------------------------------------------------------

-- lowerBits(x, divide)
-- returns the lowest bits of an integer
-- divide is the is the value that fixes the limit between higher bits and lower bits
-- examples:
--  lowerBits(x, #10) gets the lower quartet (4 bits) of byte x (8 bits)
--  lowerBits(x, #100) gets the lower bytes (8 bits) of word x (16 bits)
--  lowerBits(x, #10000) gets the lower word (16 bits) of dword x (32 bits)
public function lowerBits(integer x, integer divide)
  return remainder(x, divide)
end function

-------------------------------------------------------------------------------

-- bytes2uint(sequence, direction)
-- converts sequence of bytes peeked in memory to a numerical value
-- direction is NORMAL or REVERSE
-- when direction is REVERSE (usual), lowest byte is first
-- when direction is NORMAL (telecom frames), highest byte is first
-- example:
--   bytes2uint({#1C,#0A,#00,#00}, NORMAL) = #1C0A0000
--   bytes2uint({#1C,#0A,#00,#00}, REVERSE) = #A1C
public function bytes2uint(sequence s, integer ordering=REVERSE)
  if ordering = NORMAL then
    return bytes_to_int(reverse(s))
  elsif ordering = REVERSE then
    return bytes_to_int(s)
  end if
end function
/*
public function bytes2uint(sequence s, integer direction)
  atom l, mult

  l = 0
  mult = 1
  if direction = NORMAL then
    for i = length(s) to 1 by -1 do
      l += s[i]*mult
      mult *= #100
    end for
  elsif direction = REVERSE then
    for i = 1 to length(s) do
      l += s[i]*mult
      mult *= #100
    end for
  end if
  return l
end function
*/
-------------------------------------------------------------------------------

-- uint2bytes(integer, digits, direction)
-- converts an integer to a sequence of bytes
-- digits is the length of the resulting sequence
-- direction is NORMAL or REVERSE
-- when direction is REVERSE (usual), lowest byte is first
-- when direction is NORMAL (telecom frames), highest byte is first
-- example:
--   uint2bytes(#1C0A0000, NORMAL) = {#1C, #0A, #00, #00}
--   uint2bytes(#1C0A0000, REVERSE) = {#00, #00, #0A, #1C}
public function uint2bytes(integer n, integer digits, integer direction)
  sequence s

  s = repeat(0, digits)
  if direction = NORMAL then
    for i = digits to 1 by -1 do
      s[i] = remainder(n, #100)
      n = floor(n / #100)
    end for
  elsif direction = REVERSE then
    for i = 1 to digits do
      s[i] = remainder(n, #100)
      n = floor(n / #100)
    end for
  end if
  return s
end function

------------------------------------------------------------------------------

public function val(sequence str)
-- returns numerical value of string str
-- decimal separator may be a comma
-- displays a warning on error
  integer n
--  sequence s

  if length(str) = 0 then return 0 end if
  n = find(',', str)
  if n>0 then str[n] = '.' end if
  return to_number(str)
/*
  s = value(str)
  if s[1] = GET_SUCCESS then
    return s[2]
  else
    warnError(sprintf("Erreur de conversion: %s", {str}), 1)
  end if
*/
end function

------------------------------------------------------------------------------

public function atom2string(atom val, sequence fmt)
-- converts numerical value val according to format fmt
-- replaces decimal separator by a comma (for non english speaking countries)
  integer n
  sequence s

  s = sprintf(fmt, val)
  n = find('.', s)
  if n>0 then s[n] = ',' end if
  return s
end function

------------------------------------------------------------------------------

public function object2string(object s)
-- converts an object in a sequence
-- sub-sequences appear between curly brackets as with print()
-- sub-sequence items are separated by commas as with print()
  sequence res

  if integer(s) then
    res = sprintf("%d", s)
  elsif atom(s) then
    if (s >= 0) and (s = floor(s)) then
      res = sprintf("%.0f", s)
    elsif (s < 0) and (s = floor(s+1)) then
      res = sprintf("%.0f", s)
    else
      res = sprintf("%f", s)
    end if
  elsif string(s) then
    res = s
  else
    res = {123}  -- {
    for i=1 to length(s) do
      res &= object2string(s[i])&","
    end for
    res[$] = 125   -- }
  end if
  return res
end function

------------------------------------------------------------------------------

public function utf8ToAscii(sequence s)
  integer i, lg, byte, bytes
  sequence res
  atom char

  i = 1
  res = ""
  lg = length(s)
  if lg < 2 then return s end if
  while i <= lg do
    bytes = 1
    char = s[i]
    if and_bits(s[i],#80) = #00 then
      char = and_bits(s[i], #7F)
    elsif ((i+1) <= lg)  and   -- else Unexpected end of UTF-8 string
          (and_bits(s[i],  #E0) = #C0) and
          (and_bits(s[i+1],#C0) = #80) then
      if and_bits(s[i],#FE) != #C0 then   -- else Overlong UTF-8 char
        bytes = 2
        char = and_bits(s[i], #1F)
      end if
    elsif ((i+2) <= lg)  and   -- else Unexpected end of UTF-8 string
          (and_bits(s[i],  #F0) = #E0) and
          (and_bits(s[i+1],#C0) = #80) and
          (and_bits(s[i+2],#C0) = #80) then
      if (and_bits(s[i],  #FF) != #E0) or
           (and_bits(s[i+1],#E0) != #80) then   -- else Overlong UTF-8 char
        bytes = 3
        char = and_bits(s[i], #0F)
      end if
    elsif ((i+3) <= lg)  and   -- else Unexpected end of UTF-8 string
          (and_bits(s[i],  #F8) = #F0) and
          (and_bits(s[i+1],#C0) = #80) and
          (and_bits(s[i+2],#C0) = #80) and
          (and_bits(s[i+3],#C0) = #80) then
      if (and_bits(s[i],  #FF) != #F0) or
           (and_bits(s[i+1],#F0) != #80) then   -- else Overlong UTF-8 char
        bytes = 4
        char = and_bits(s[i], #07)
      end if
    elsif ((i+4) <= lg)  and   -- else Unexpected end of UTF-8 string
          (and_bits(s[i],  #FC) = #F8) and
          (and_bits(s[i+1],#C0) = #80) and
          (and_bits(s[i+2],#C0) = #80) and
          (and_bits(s[i+3],#C0) = #80) and
          (and_bits(s[i+4],#C0) = #80) then
      if (and_bits(s[i],  #FF) != #F8) or
         (and_bits(s[i+1],#F8) != #80) then   -- else Overlong UTF-8 char
        bytes = 5
        char = and_bits(s[i], #03)
      end if
    elsif ((i+5) <= lg) and  -- else Unexpected end of UTF-8 string
          (and_bits(s[i],  #FE) = #FC) and
          (and_bits(s[i+1],#C0) = #80) and
          (and_bits(s[i+2],#C0) = #80) and
          (and_bits(s[i+3],#C0) = #80) and
          (and_bits(s[i+4],#C0) = #80) and
          (and_bits(s[i+5],#C0) = #80) then
      if (and_bits(s[i],  #FF) != #FC) or
         (and_bits(s[i+1],#FC) != #80) then   -- else Overlong UTF-8 char
        bytes = 6
        char = and_bits(s[i], #01)
      end if
    end if
    if bytes > 1 then
      for n=1 to bytes-1 do
        byte = and_bits(s[i+n],#3F)
        char = (char*64)+byte
      end for
      if ((char>=#D800) and (char<=#DFFF)) or
         (remainder(char,#100)>=#FE) then
        res &= s[i]
        for n=1 to bytes-1 do
          res &= s[i+n]
        end for
      else
        res &= char
      end if
      i += bytes-1
    else
      res &= char
    end if
    i += 1
  end while
  return res
end function

------------------------------------------------------------------------------

public function hexString(sequence s)
-- returns hexadecimal representation of a sequence
  sequence result

  result = ""
  for i = 1 to length(s) do
    result &= sprintf("%02x", s[i])
  end for
  return result
end function

------------------------------------------------------------------------------

public function dos2unix(sequence s)
-- converts DOS end of lines to Unix ones
  sequence t
  integer l

  t = {}
  l = match({13,10}, s)
  while l > 0 do
    t = t & s[1..l-1]
    s = s[l+1..length(s)]
    l = match({13,10}, s)
  end while
  t = t & s
  return t
end function

